#!/usr/bin/perl -Tw
#-----------------------------------------------------------------------------
# webitrans.pl
#
# Web interface to "itrans"
# This script is similar to the script used for the Online Interface at:
# http://www.aczone.com/itrans/online/
# and is provided as an example. There is no documention for this script,
# other than program comments in this file itself.
# -----------------------------------------------------------------------
# This program accepts www web form input using ITRANS encoded-English,
# and produces printouts in Indian Language Scripts.
#-----------------------------------------------------------------------------
my $VERSION = "1.21";
# Last Modified: July 18, 2001
# Major updates by Avinash Chopde <avinash@acm.org> http://www.aczone.com/
#-----------------------------------------------------------------------------
# Based on script by Hari Adiseshu
# - changes by Atul Narkhede (for alkhemy.com website)
# - changes by shree on 3/2/2000
# - changes Shashikant Joshi 03/10/2000
# - changes by Nandu
# -----------------------------------------------------------------------
# March 2001, Avinash Chopde <avinash@acm.org> http://www.aczone.com/
# Many changes.
# --> Accept file upload also (multipart forms).
# --> .ps to .gif transformation modified to create anti-aliased text.
# --> more fonts added, more sizes added.
# --> two formats added 
#     "itrans-itx" for handling .itx files directly
#     "isongs-s" for handling .s ISONGS files directly
# --> HTML output in both ISO-8859-1 (Latin1) and Unicode UTF-8
#------------------------------
# Rough outline of things needed to make this work at your site:
# Programs needed, on Linux: TeX (teTeX)
# GhostScript - use version 5.10 - it is twice as fast as version 6.5
# netpbm - using version 9.14 (May 2001)
# itrans - install itrans 5.3+, normally goes in /usr/local/itransNN
# CGI.pm - latest one - 2.752 or newer
# ----
# Following steps outline things I had to do to get my Linux (Slackware)
# system up and running.
# I keep all my CGI files under the /home/cgi folder.
# So, the ROOTDIR for web interface to itrans is at /home/cgi/itrans.
# /home/cgi/ also contains two other folders: tmp and bin, used by all
# CGI tools I serve from this Linux box.
#
# Make this folder non-writeable:
# $ROOTDIR
# Safest is to make all files and directories non-writeable,
# and all directories below $ROOTDIR should be additionally non-readable
# by group and others (ROOTDIR itself needs to be readable, Apache reads
# it to find index.html, index.shtml, etc).
# commands: chmod -R a-w <ITRANS/ROOTDIR>; chmod go-r <all subdirectories>
# 
# Files/folders in $ROOTDIR:
#   commonWebInt.txt
#   index.html
#  Template files, defined in script:
#   $WEBWAIT_THTML
#   $WEBDONE_THTML
#   $TEX_BASE
#   $LAT1_BASE
#   $UTF8_BASE
#   $TEX_BASE_INC
#   $LAT1_BASE_INC
#   $UTF8_BASE_INC
#
#   I symlink /home/cgi to Apache htdocs, so http:/cgi/itrans/ will get
#   above listed index.html
# Additional  files in ROOTDIR:
#   prepi2h.pl
# Files in http:/cgi-bin/
#   webitrans.pl     [I symlink /home/cgi/itrans/webitrans.pl to this]
# ----
#   /home/cgi/tmp --> Make this folder writeable by everybody
#
# Config files:
#   /home/cgi/itrans/dvipsrc
# ---
# Finally, edit the top section of this CGI script to point to files at
# your site.
# To test, run with the -t option, using the example webtest.txt or similar
# files:
# webitrans.cgi -t /tmp/webtest.txt
# -t <filename> should use full path for filename.
# --
# This script configures TeX/LaTeX to never run mktextfm, mktekpk when
# running under the web - so, an assumption is made that each
# web admininstrator will create all required TFM/PK files (use the
# test mode -t webtest.txt of this script, and then manually run latex/dvips
# on the created .tex file).
# Of course, this is to the discretion of the web admin - you can decide
# whether to turn this on or off -- see the LATEX_E/DVIPS_E vars below.
# -- Make dvips use 300 dpi fonts - jetiiisi mode, makes for much smaller
# .ps files than if 600 dpi/ljfour is used!
# -- Hardest thing to do on my machine was getting teTeX to put *pk
# fonts in the right place! Could not get that work - I wanted
# fonts to go in /var/cache/fonts, but mktexpk always put ITRANS fonts
# in current directory.
# dvips/mktexpk did put the cmr*pk fonts in /var/cache/fonts, it only
# messed up on dvng, tel, kan, wntml fonts. BUT - even the cmr*pk files
# were not handled correctly - I had enabled "nomode" (don't use printer
# name subdir for pk files), while mktexpk honored that, dvips still
# looked for that folder to search for cmr*pk files - which did not work
# of course! So, each time I ran dvips, it wanted to ceated the cmr*pk files!
# kpathsea in teTeX and texconfig has problems - don't disable
# "mode" from "Font directories -> Options" - it will mess up a lot!
# Don't change the defaults, they cause more trouble than it is worth.
# I run dvips in -M mode (don't run maketexpk), so I pre-created all
# the required *pk files. These will be left in the current dir.
# I then copied them to the /var/cache/fonts/pk/jetiiisi/other folder
# [I created the "other" folder in the standard /var/cache/fonts/pk/jetiiisi
# folder]
# Then, run "mktexlsr" (this will update the /var/cache/fonts/ls-R file.)
# IN SHORT: 
# 1: run itrans/latex/dvips on text file, using all possible fonts of the
#    web interface.
# 2: this will create some *pk files in current folder.
#    mkdir /var/cache/fonts/pk/jetiiisi/other
#    mv *pk /var/cache/fonts/pk/jetiiisi/other
# 3: update ls-lr
#    mktexlsr
#----
# Install latex2html (latest version) 
# http://saftsack.fs.uni-bayreuth.de/~latex2ht/current/
# This is needed just to get correct html.sty, so if any user
# sends any ITRANS doc/*.itx file, it will be processed correctly
# by this Online ITRANS interface.
# latext2html is also useful for the pstoimg script - it can create
# multipage GIF output:
# pstoimg -multipage -aaliastext -crop tblr -density 100 -white -transparent -type gif -out <output>.gif <input>.ps
#----
# TeX security - texmf.cnf file (in /usr/share/texmf/web2c)
# turn off shell_escape \write18,
# restrict input, openin, openout - how to do it for stuff under
# this perl script only? Don't know - for now, edit texmf.cnf
# and set openin_any = p (to paranoid) and openout_any p (to paranoid)
# and shell_escape = f  (write18 not allowed).
# May 2001: Only openout set to paranoid for now, openin left as is...
#----
# Apache: turn off Indexes directive for all Directories to prevent directory
# listings (since tmp directories here are readable by all, this is done
# to prevent casual users from seeing file names in some other tmp folder).
#----
#
#-----------------------------------------------------------------------------
use 5.005; # perl newer than 5.005 required
use CGI 2.50 qw(escapeHTML); # version 2.50 for Vars()
use File::Copy;
use File::Basename;
use Cwd;
use Getopt::Std;
use POSIX qw(floor);
use Socket qw(:DEFAULT :crlf);
#-----------------------------------------------------------------------------
$start_time = time();

{ # was sub BEGIN, but looks like I cannot call subroutines in here...
    # HOMEDIR is a temporary variable, used in setting all the global
    # variables used by this script. You don't have to use this or set it.
  $HOMEDIR="/home/cgi"; # contains tmp/  bin/  itrans/  cdinsert/  etc

    # log file - global script issues, not related to any one invocation
    # make sure this file exists, and is chmod 662 - writeable, but not readable
    # by others so casual web hackers can't read this directly using a browser.
  $LOGFILE = "$HOMEDIR/tmp/weblog.txt";

  $CGI::POST_MAX=1024 * 500;  # max size posts accepted, bytes

  &initialize();
}

$SIG{HUP} = $SIG{INT} = $SIG{QUIT} = $SIG{PIPE} = $SIG{TERM} = \&sighandler;
#-----------------------------------------------------------------------------
$DEBUG = 1; # 0 no debug messages, 1 some messages, 2 more.
# Debug messages may go to Apache error_log at the beginning, but
# once the LOGFILE/MSGFILE is created, they will go there.

#----------------------------
# Following vars need to be set specifically for each site 
$ITRANSDIR = "/usr/local/itrans53"; # where itrans has been installed

# ROOTDIR dir is the folder which Apache uses for http:/$ROOTHTTP (locally)
# "/usr/local/apache/htdocs/cgi" is symlink to "/home/cgi" on my machine, so:
$ROOTDIR = "$HOMEDIR/itrans"; # contains html files

$ENV{'PATH'}="/bin:/usr/bin:/usr/local/bin:/usr/share/texmf/bin:/usr/local/netpbm/bin:$HOMEDIR/bin";
# security blanket - make sure all folders/files are non-writeable by others!

# create temp folder name
# $TDATE = sprintf("%02d%02d%02d", (localtime($start_time))[3], (localtime($start_time))[2], (localtime($start_time))[1]); # current date hour minute
$TDATE = sprintf("%02d%02d", (localtime($start_time))[3], (localtime($start_time))[2]); # current date hour
$WORKDIRNAME = "it" . $TDATE;
$WORKDIRNAME = &mktempdir("$HOMEDIR/tmp", $WORKDIRNAME, floor(rand(1e3))); # 1e3 -> max 3 digits
$WORKDIR = "$HOMEDIR/tmp/$WORKDIRNAME"; 
$WORKHTTP = "/cgi/tmp/$WORKDIRNAME"; 

# for each invocation - semi-random name for files - gets some privacy, since
# these folders and files are readable by the world.
$WORKID = floor(rand(1e4)); # max 4 digits
$WORKTEX = "tex" . $WORKID; # file name to use for intermediate files
$WORKLAT1 = "lat1" . $WORKID; # file name to use for intermediate files
$WORKUTF8 = "utf8" . $WORKID; # file name to use for intermediate files

$MSGFILENAME = "log$WORKID.txt";
$MSGFILE = "$WORKDIR/$MSGFILENAME";
# STDOUT/STDERR messages collected here, different file for each invocation

$SAVINPUT="input.txt"; # file exact copy of form input string, saved for user

# input templates, include files: all these files reside in $ROOTDIR
$WEBWAIT_THTML = "$ROOTDIR/wait_t.html"; # partial HTML file, template
$WEBDONE_THTML = "$ROOTDIR/done_t.html"; # partial HTML file, template
$WEBDONE_HTML = "done$WORKID.html"; # complete HTML file (created in $WORKDIR)

$TEX_BASE = "$ROOTDIR/itrtex.itx"; # itrans input, latex mode
$LAT1_BASE = "$ROOTDIR/itrlat1.itx"; # itrans input, html latin1 mode
$UTF8_BASE = "$ROOTDIR/itrutf8.itx"; # itrans input, html unicode mode

$TEX_BASE_INC="itrtex.inc"; # $TEX_BASE will have line #include=<this file>
$LAT1_BASE_INC="itrlat1.inc"; # $LAT1_BASE will have line #include=<file>
$UTF8_BASE_INC="itrutf8.inc"; # $UTF8_BASE will have line #include=<file>

# following may be left alone, in most cases
$ENV{'SHELL'} = "/bin/sh";
$ENV{'TMPDIR'} = $WORKDIR;
$ENV{'TEMP'} = $WORKDIR;
$ENV{'TZ'} = "EST5EDT";

$ENV{'DVIPSRC'} = "$ROOTDIR/dvipsrc";
$ENV{'TEXINPUTS'} = ":$ITRANSDIR/lib:$ITRANSDIR/lib/fonts:$ROOTDIR";
$ENV{'TEXFONTS'} = ":$ITRANSDIR/lib/fonts";
$ENV{'MFINPUTS'} = ":$ITRANSDIR/lib/fonts";

delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};   # Make %ENV safer

# constants ----------------------------------------------------------

# Each language/script supported:
@SCRIPTS = (
    'gujarati',
    'bengali',
    'marathi',
    'hindi',
    'sanskrit',
    'tamil',
    'telugu',
    'kannada',
    'gurmukhi',
    'roman',
);

# Each text size supported, array holds HTML output to be produced
%FONTSIZES = (
    'small', 'SIZE="-1"',
    'normal', 'SIZE="+0"',
    'large', 'SIZE="+2"',
    'huge', 'SIZE="+4"',
);

$FORMATPROSE = "itrprose"; # leave line breaking to TeX, or Web Browser, etc
$FORMATITX = "itrans-itx"; # itrans .itx file provided as input
$FORMATISONGS = "isongs-s"; # itrans song book .s file provided as input

# --- MRKR_* variables are strings found in input HTML template files,
# to be replaced with job-specific values
$MRKR_HOSTNAME = "MRKR_HOSTNAME";
$MRKR_WEBDONEHTML = "MRKR_WEBDONEHTML";
$MRKR_PDF = "MRKR_PDF";
$MRKR_POSTSCRIPT = "MRKR_POSTSCRIPT";
$MRKR_GIF = "MRKR_GIF";
$MRKR_HTML_LATIN1 = "MRKR_HTML_LATIN1";
$MRKR_HTML_UTF8 = "MRKR_HTML_UTF8";
$MRKR_INPUT = "MRKR_INPUT";
$MRKR_SHOWSTATUS = "MRKR_SHOWSTATUS";
$MRKR_ITRANSERROR = "MRKR_ITRANSERROR";
$MRKR_GIFERROR = "MRKR_GIFERROR";
$MRKR_SYSERROR = "MRKR_SYSERROR";
$MRKR_SUCCESS = "MRKR_SUCCESS";
$MRKR_ANYERROR = "MRKR_ANYERROR";
$MRKR_LATEXGIF = "MRKR_LATEXGIF";
$MRKR_MSGFILE = "MRKR_MSGFILE";
$MRKR_TIMETAKEN = "MRKR_TIMETAKEN";
$MRKR_DONTREDIRECT = "MRKR_DONTREDIRECT";

#--- tag printed in the HTML file, denotes start and end of user entered text
# must be a HTML comment

$HTML_TAG_USER_TEXT = "<!-- User Text -->";

#--- executables - since am setting PATH, no need to give full path for each

$REDIRECT="< /dev/null 2>&1";

# -- cmd 0:
$ITRANS_E = "itrans -i $WORKTEX.itx -o $WORKTEX.tex $REDIRECT"; # tex 
# -- cmd 1:
$LATEX_E = "latex -no-mktex=tfm $WORKTEX.tex $REDIRECT"; # don't run mktextfm
# -- cmd 2:
$DVIPS_E = "dvips -R -M $WORKTEX.dvi -o $WORKTEX.ps $REDIRECT"; # don't run mktexpk (-M), and run in secure mode (-R)
# -- cmd 3:
$PS2PGM_E = "gs -q -dNOPAUSE -dBATCH -dTextAlphaBits=4 -dGraphicsAlphaBits=4 -r100 -sDEVICE=pgmraw -sOutputFile=$WORKTEX.pgm -f $WORKTEX.ps $REDIRECT";
# gs options:
#   - made it anti-alias text (-dTextAlphaBits)
#   - use 100dpi resolution (default is 72, I think)
#   - pgm (gray scale) is faster and 66% smaller file than ppm (full color)
#     pgmraw is even more smaller, and is 10 times faster too, for large files!
#   ? doing an initial 'save' helps keep fonts from being flushed between pages.
# -- cmd 4:
$PGM2GIF_E = "(pnmcrop $WORKTEX.pgm | pnmpad -white -l5 -r5 -t5 -b5 | ppmtogif -transparent white -interlace) 2>&1 > $WORKTEX.gif";
# 2>&1 and then > file makes STDERR go to old STDOUT, and new STDOUT to file
# -- cmd 5:
$PS2PDF_E = "ps2pdf $WORKTEX.ps $REDIRECT";

# HTML output commands

# UTF-8 charset
$ITRANSU_E = "itrans -U -i $WORKUTF8.itx -o $WORKUTF8.htm $REDIRECT"; # html UTF-8

# user-defined charset (fake ISO-Latin1 fonts displaying Indian Languages)
$PREPI2H_E = "$ROOTDIR/prepi2h.pl $LAT1_BASE_INC > tmp$LAT1_BASE_INC $REDIRECT && mv tmp$LAT1_BASE_INC $LAT1_BASE_INC $REDIRECT"; # html
$ITRANSH_E = "itrans -8 -i $WORKLAT1.itx -o $WORKLAT1.htm $REDIRECT"; # html

@ALLCMDS = ($ITRANS_E, $LATEX_E, $DVIPS_E, $PS2PDF_E, $PS2PGM_E, $PGM2GIF_E);
@ALLHTMLCMDS = ($ITRANSU_E, $PREPI2H_E, $ITRANSH_E);

#-----------------------------------------------------------------------------
# initialize 

$time_taken = 0;
$hostname = "";
$useragent = "";
$status_count = -2;
$errflag = 0;
$errcmd = -1;
$dontredirect = ""; # make this 1 to prevent automatic redirect in WEBWAIT_THTML

chdir $WORKDIR or myexit("cd ($WORKDIR) failed: $!");

open(MSGFILE, ">> $MSGFILE")  
  or myexit("Could not open message file ($MSGFILE): $!");
select((select(MSGFILE), $| = 1)[0]); # set to autoflush

open(WEBWAIT_THTML, "< $WEBWAIT_THTML")  
  or myexit("Could not open WEBWAIT_THTML file ($WEBWAIT_THTML): $!");

open(WEBDONE_THTML, "< $WEBDONE_THTML")
  or myexit("Could not open WEBDONE_THTML file ($WEBDONE_THTML): $!");

open(WEBDONE_HTML, "> $WORKDIR/$WEBDONE_HTML")
  or myexit("Could not open WEBDONE_HTML file ($WORKDIR/$WEBDONE_HTML): $!");

$datestr = localtime($start_time);
print MSGFILE "[$datestr]: Starting job in directory $WORKDIRNAME\n";

# ---- file WEBWAIT opened, copy it to output, piece by piece, showing status
# -- this relies on the client side of things - the web browser - to display
# the page even though it is not complete. Most browsers do this - IE on PC,
# Netscape on PC and Mac. But IE on Mac waits until it sees the last </BODY>
# so this does not work well in all cases... 

# start_html etc is not used - instead, a template HTML file is read
# in and output to STDOUT
&print_cgihtml($MRKR_SHOWSTATUS); # keep copying until MRKR_SHOWSTATUS line...
print &show_status();

#----------------------------------------------------------------
# collect user input

$query = new CGI;

getopt('t:'); # -t <filename> is test mode. Filename is used as $instring

if (! $opt_t ) {
    $hostname = $query->remote_host();
    $useragent = $query->user_agent();
    $inlanguage = $query->param('language');
    $infontsize = $query->param('fontsize');
    $informat = $query->param('format');
    $inoutput = $query->param('output');
    $instring = $query->param('string');
    $infile = $query->param('filename');
} else {
    $hostname = qq(testing mode "-t $opt_t");
    $useragent = "";
    $inlanguage = 'sanskrit';
    $infontsize = 'normal';
    $informat = 'itrlines';
    $infile = '';
    $inoutput = '';
    open(INPUT, "$opt_t")
      or myexit("Could not open input file (-t $opt_t) [Make sure full path given -t /tmp/x.txt etc]: $!");
    print MSGFILE "opened input file $opt_t\n" if ($DEBUG >= 1);
    while (<INPUT>) {
	# read each line to get correct EOLN value for this platform (works?)
	s/$CR?$LF/\n/; # variables from Socket package
        $instring .= $_; 
    }
    close INPUT;
}

print LOGFILE "$WORKDIRNAME itrans [$datestr] $hostname - started\n";

$html_only = ($inoutput eq 'itrhtmlonly'); # html latin1 and utf8 output only...

# check for CGI error - post too big, or user hit STOP, etc...
$cgierror = $query->cgi_error();
if ($cgierror) {
    myexit($cgierror);
}

print MSGFILE "----------------------------------------------------\n";
print MSGFILE "Got these values from the form:\n";
my %params = $query->Vars();
while (($key, $value) = each %params) {
    # assuming value is single string - if multi-valued, need
    # to split on \0 to get array of values...
    $value = "<see file $SAVINPUT>" if ( $key =~ /^string$/ );
    print MSGFILE "  $key = '$value'\n";
}
print MSGFILE "Some environment vars:\n";
print MSGFILE "  remote_host = '$hostname'\n";
print MSGFILE "  user_agent = '$useragent'\n";
print MSGFILE "----------------------------------------------------\n";

$gotstring = ($instring =~ /\S+/);
$gotfile = ($infile =~ /\S+/);

($gotstring || $gotfile) or myexit("Empty input - no input text or file found!");

# save users string in file, in case they need to use it again
open(OUTPUT, "> $SAVINPUT")
  or myexit("Could not open output file to save form input ($SAVINPUT): $!");
print MSGFILE "opened file $SAVINPUT to save input text\n" if ($DEBUG >= 1);

if ($gotfile) { # ignore $gotstring, file takes precedence
    if ($gotstring) {
	print MSGFILE "** Warning: user entered text as well as filename, ignoring text.\n",
    }
    while (<$infile>) {
	# read each line to get correct EOLN value for this platform (works?)
	s/$CR?$LF/\n/; # variables from Socket package
	print OUTPUT $_;
    }

    print MSGFILE "... read in uploaded file: $infile\n" if ($DEBUG >= 1);

    if ($infile =~ /\.itx$/ && $informat !~ /^$FORMATITX$/) {
        # file name ends in .itx, but informat is wrong...
        $informat = $FORMATITX; # force format 
	print MSGFILE "** Warning: user sent .itx file, but wrong page format, changed to $FORMATITX.\n";
    }
    if ($infile =~ /\.s$/ && $informat !~ /^$FORMATISONGS$/) {
        # file name ends in .s, but informat is wrong...
        $informat = $FORMATISONGS; # force format 
	print MSGFILE "** Warning: user sent .s file, but wrong page format, changed to $FORMATISONGS.\n";
    }
} else {
    # fix for correct EOLN value for this platform (works?)
    $instring =~ s/$CR?$LF/\n/g; # variables from Socket package
    print OUTPUT  $instring;
}
close OUTPUT;

#-----------------------------------------------------------------------------
# two formats need to be treated specially - .itx files and .s files
# all other formats are web-input text formats, containing commands
# like {script-size} that need to be replaced by ITRANS commands

# LaTeX pre-processing....
if ($informat =~ /^$FORMATITX$/) {
    # format is $FORMATITX, use the user supplied file or text directly
    copy($SAVINPUT, "$WORKTEX.itx")
      or myexit("Could not copy ($SAVINPUT) to ($WORKTEX.itx): $!");
} else {
    # handle all other formats, including $FORMATISONGS
    &preprocess_itrinput();
}

# HTML output pre-processing....
if ($informat =~ /^$FORMATITX$/) {
    # format is $FORMATITX, use the user supplied file or text directly
    copy($SAVINPUT, "$WORKLAT1.itx")
      or myexit("Could not copy ($SAVINPUT) to ($WORKLAT1.itx): $!");
    copy($SAVINPUT, "$WORKUTF8.itx")
      or myexit("Could not copy ($SAVINPUT) to ($WORKUTF8.itx): $!");
} else {
    # handle all other formats, including $FORMATISONGS
    &preprocess_htminput();
}

# now have $WORKTEX.itx and $WORKLAT1/UTF8.itx
# got all data in ITRANS format, now run all the programs to finally
# get the .gif and .ps files, as well as HTML output

$datestr = localtime(time());
print MSGFILE "----------------------------------------------------\n";
print MSGFILE "[$datestr] $hostname :: Starting programs...\n";
print MSGFILE "----------------------------------------------------\n";

#-----------------------------------------------------------------------------
# first, create HTML output

print MSGFILE "-------  Creating HTML Output - UTF-8 and Latin1 ---\n";
print MSGFILE "----------------------------------------------------\n";

foreach $i (0 .. $#ALLHTMLCMDS) {
    my ($t1, $td);

    $cmd = $ALLHTMLCMDS[$i];
    print &show_status($cmd); # status messages to web 0 .. 1 .. etc
    $t1 = time();
    $out = `$cmd`;
    $td = time() - $t1;
    $returncode = ($? >> 8); 

    print MSGFILE qq("$cmd" executed\n -- Took $td seconds, returns $returncode\n);
    print MSGFILE "$out\n---------------------\n";

    if ($returncode != 0) {
	if ($html_only) {
	    myexit("ITRANS failed - HTML output mode.");
        } else {
	    print "<BR>\n";
	    print "Session ID# <B>$WORKDIRNAME</B> - Warning - HTML output not produced\n<BR>\n";
	    print "Warning: problem executing \"$cmd\", will continue anyway.\n";
	    print "<P>\n";
	}
	# last; # no last continue - try other commands - UTF8/Latin1/etc
    }
}

# add <BR> if needed to the  output html files
$returncode = &postprocess_htm();
if ($returncode != 0) {
    if ($html_only) {
	myexit("ITRANS post-processing failed - HTML output mode.");
    } else {
	print "<BR>\n";
	print "Session ID# <B>$WORKDIRNAME</B> - Warning - HTML output not produced correctly\n<BR>\n";
	print "Warning: problem post-processing HTML files, will continue anyway.\n";
	print "<P>\n";
    }
}

if (! $html_only) {
    #------------------------------------------------------------------------
    # second, run all the programs to get GIF/PS etc output, LaTeX processing
    print MSGFILE "-------  Creating GIF/PS/PDF Output ----------------\n";
    print MSGFILE "----------------------------------------------------\n";

    foreach $i (0 .. $#ALLCMDS) {
	my ($t1, $td);

	$cmd = $ALLCMDS[$i];
	print &show_status($cmd); # status messages to web 0 .. 1 .. etc
	$t1 = time();
	$out = `$cmd`;
	$td = time() - $t1;
	$returncode = ($? >> 8); 

	print MSGFILE qq("$cmd" executed\n -- Took $td seconds, returns $returncode\n);
	print MSGFILE "$out\n---------------------\n";

	if ($returncode != 0) {
	    print "<P>\n";
	    print "Session ID# <B>$WORKDIRNAME</B> - an error occured, cannot proceed.\n<BR>\n";
	    # print "Command: ", escapeHTML($cmd), ", \$? is $?\n<BR>";
	    # $out = escapeHTML($out);
	    # $out =~ s/\n/\n<BR>/;
	    # print qq($out\n<BR>\n);
	    $errflag++;
	    $errcmd = $i;
	    last;
	}
    }
}

print " Done!\n<P>\n"; # status messages to web...

print MSGFILE "commands executed. \$errflag='$errflag' \$errcmd='$errcmd'\n";

# remove intermediate log files
unlink("$WORKTEX.tex")
   unless ($errflag || $opt_t || $DEBUG >= 1);
unlink("$WORKTEX.aux", "$WORKTEX.dvi", "$WORKTEX.log", "$WORKTEX.pgm")
   unless ($opt_t || $DEBUG >= 2);

$end_time = time();
$time_taken = $end_time - $start_time;
$datestr = localtime($end_time);
$datestr = "$WORKDIRNAME itrans [$datestr] took $time_taken secs " .
           ($errflag ? "[error]" : "[no error]") . "\n";
print LOGFILE $datestr;
print MSGFILE $datestr;
print MSGFILE "----------------------------------------------------\n";

# copy output HTML page show output or error, as appropriate

while (<WEBDONE_THTML>) {

    if ($errflag) {
      # depending on error, uncomment particular lines in WEBDONE_THTML
      next if (&check_marker($_, $MRKR_ANYERROR));
      if ($ALLCMDS[$errcmd] =~ /$ITRANS_E/) {
	next if (&check_marker($_, $MRKR_ITRANSERROR));
      } elsif ($ALLCMDS[$errcmd] =~ /$PGM2GIF_E/) {
	next if (&check_marker($_, $MRKR_GIFERROR));
      }
      next if (&check_marker($_, $MRKR_SYSERROR));
      # $errcmd not ITRANS or PGMGIF
    } else { 
      # no error, uncomment the SUCCESS lines
      next if (&check_marker($_, $MRKR_SUCCESS));
      # no error, uncomment the LATEXGIF line if latex/pdf/ps/gif produced
      next if (!$html_only && &check_marker($_, $MRKR_LATEXGIF));
    }

    print WEBDONE_HTML &replace_markers($_);
}
close(WEBDONE_HTML);

# complete rest of HTML from this CGI script, will redirect to WEBDONE
&print_cgihtml("");

close(WEBDONE_THTML);
close(WEBWAIT_THTML);
close(MSGFILE);
close(LOGFILE);

exit($errflag);

#-----------------------------------------------------------------------------
# Subroutines

sub preprocess_itrinput {
    # Pre-process the input string, to replace web commands {script-textsize}
    # into ITRANS commands.

    print MSGFILE "--- pre-processing input, for tex/gif/ps/pdf output\n";

    copy($TEX_BASE, "$WORKTEX.itx")
      or myexit("Could not copy ($TEX_BASE) to ($WORKTEX.itx): $!");

    open(OUTPUT, "> $TEX_BASE_INC")  
      or myexit("Could not open \$TEX_BASE_INC file ($TEX_BASE_INC): $!");

    open(INPUT, "< $SAVINPUT")
      or myexit("Could not re-open saved input string file ($SAVINPUT)]: $!");
    print MSGFILE "Re-opened input file $SAVINPUT\n" if ($DEBUG >= 1);

    # begin: pre-pend the default script/size
    if ($informat =~ /^$FORMATISONGS$/) {
	print OUTPUT  "\\itrsize$infontsize\\itrfonteng ";
    } else {
	print OUTPUT  "\\begin{$informat}\n\\itrsize$infontsize\\itrfonteng #$inlanguage ";
    }

    $i = 0;
    while (<INPUT>) {
      my $line = $_;
      $i++;
      my %scriptsizes = &find_tags($line, $i);

      # now replace all script-size commands
      while ( ($key) = each %scriptsizes) { # just $key, don't need $value
	$key =~ /(\w+)-(\w+)/;
	# $replace = "##\\itrsize$2";  # end prev language, set font size
	# $replace .= "\\itrfonteng #$1 "; # set english font, start new script
	$script = $1;
	$size = $2;
	$line =~ s/\{$script-$size\}/##\\itrsize$size\\itrfonteng #$script /g;
	print MSGFILE "$0: line $i found key $key replacing ($script-$size)\n" if ($DEBUG >= 2);
      }

      print OUTPUT $line;
    } # end while <INPUT>

    # end: append the end<script> ITRANS commands
    if ($informat =~ /^$FORMATISONGS$/) {
	print OUTPUT  "\n";
    } else {
	print OUTPUT  "##\\end{$informat}\n";
    }

    close(INPUT);
    close(OUTPUT);

    print MSGFILE "--> input text pre-processed and copied to $TEX_BASE_INC\n"
      if ($DEBUG >= 1);
}

#----------------------------------------
sub preprocess_htminput {
    # Pre-process the input string, to replace web commands {script-textsize}
    # into ITRANS commands.
    my ($str);
    local (*OUTPUT, *INPUT);

    print MSGFILE "--- pre-processing input, for html output\n";

    copy($LAT1_BASE, "$WORKLAT1.itx")
      or myexit("Could not copy ($LAT1_BASE) to ($WORKLAT1.itx): $!");
    copy($UTF8_BASE, "$WORKUTF8.itx")
      or myexit("Could not copy ($UTF8_BASE) to ($WORKUTF8.itx): $!");

    open(OUTPUT, "> $LAT1_BASE_INC")  
      or myexit("Could not open \$LAT1_BASE_INC file ($LAT1_BASE_INC): $!");

    print OUTPUT $HTML_TAG_USER_TEXT, "\n";

    open(INPUT, "< $SAVINPUT")
      or myexit("Could not re-open saved input string file ($SAVINPUT)]: $!");
    print MSGFILE "Re-opened input file $SAVINPUT\n" if ($DEBUG >= 1);

    # begin: pre-pend the default script/size
    if ($informat =~ /^$FORMATISONGS$/) {
	$str = "<FONT $FONTSIZES{$infontsize}>";
    } else {
	$str = "<FONT $FONTSIZES{$infontsize}>#$inlanguage ";
    }
    print OUTPUT $str;

    $i = 0;
    while (<INPUT>) {
      my $line = $_;
      $i++;
      my %scriptsizes = &find_tags($line, $i);

      # now replace all script-size commands
      while ( ($key) = each %scriptsizes) { # just $key, don't need $value
	$key =~ /(\w+)-(\w+)/;
	$script = $1;
	$size = $2;
	$line =~ s!\{$script-$size\}!##</FONT><FONT $FONTSIZES{$size}>#$script !g;
	print MSGFILE "$0: line $i found key $key replacing ($script-$size)\n" if ($DEBUG >= 2);
      }

      print OUTPUT $line;
    } # end while <INPUT>

    # end: append the end<script> ITRANS commands
    $str = ($informat =~ /^$FORMATISONGS$/) ? "</FONT>\n" : "##</FONT>\n";
    print OUTPUT $str;

    print OUTPUT $HTML_TAG_USER_TEXT, "\n";

    close(INPUT);
    close(OUTPUT);

    print MSGFILE "--> input text pre-processed and copied to $LAT1_BASE_INC & $UTF8_BASE_INC\n"
      if ($DEBUG >= 1);

    copy($LAT1_BASE_INC, $UTF8_BASE_INC)
      or myexit("Could not copy ($LAT1_BASE_INC) to ($UTF8_BASE_INC): $!");
}

#----------------------------------------
sub postprocess_htm {
    # Post-process the output HTML file, add <BR> if needed
    my ($marker, $ret);
    local (*OUTPUT, *INPUT);

    print MSGFILE "--- post-processing html output, for verse and isongs\n";
    $ret = 0;

    if ($informat =~ /^$FORMATPROSE$/ || $informat =~ /^$FORMATITX$/) {
      print MSGFILE "    no post-processing needed, is not verse or isongs\n";
      return $ret;
    }

    # append a <BR> to every line between HTML_TAG_USER_TEXT

    foreach $file ("$WORKLAT1.htm", "$WORKUTF8.htm") {

      open(INPUT, "$file")
	or (print MSGFILE "Warning: open html output file ($file): $!"), $ret++, next;

      print MSGFILE "Re-opened output file $file\n" if ($DEBUG >= 1);

      open(OUTPUT, "> $file.tmp")  
	or (print MSGFILE "Warning: open temp output file ($file.tmp): $!"), $ret++, next;

      $marker = 0; # 1 means seen first HTML_TAG_USER_TEXT, 2 means seen both

      while (<INPUT>) {
	$marker++ if (/$HTML_TAG_USER_TEXT$/);
	s/$/<BR>/ if ($marker == 1);
	print OUTPUT;
      }

      close(INPUT);
      close(OUTPUT);

      move("$file.tmp", "$file")
        or (print MSGFILE "Warning: rename temp file ($file.tmp to $file): $!"), $ret++, next;

   }

   print MSGFILE "---------------------\n";

   return $ret;
}

#----------------------------------------
sub show_status {
    my($cmd) = @_;

    $status_count++;
    if ($status_count < 0) { return ("Working . .<BR>\n"); }

    # if cmd contains | char, look for last | and the command after that
    # otherwise, use the first word as the command
    # command can be any combination of \w / \ . characters.
    ($cmd =~ /\|\s+([\\\/\.\w]+)[^|]*$/) || ($cmd =~ /\s*([\\\/\.\w]+)/); # $1 is last command.
    if ($1) {
	$cmd = basename($1) || $1;
        return "&nbsp;&nbsp; ($cmd) . .<BR>\n";
    }

    return "&nbsp;&nbsp; $status_count . .<BR>\n";
}

#----------------------------------------
sub check_marker {
    # check if $str contains the given marker string (if $in marker != "")
    my($str, $in) = @_;
    $in && $str =~ /\b$in\b/;
}

#----------------------------------------
sub replace_one {
    # replaces $str in calling function (if $in marker != "")
    my($str, $in, $out) = @_;
    $in && $_[0] =~ s/\b$in\b/$out/;
}

#----------------------------------------
sub replace_markers {
    my($str) = @_;
    &replace_one($str, $MRKR_HOSTNAME, "$hostname");
    &replace_one($str, $MRKR_WEBDONEHTML, "$WORKHTTP/$WEBDONE_HTML");
    &replace_one($str, $MRKR_INPUT, "$WORKHTTP/$SAVINPUT");
    &replace_one($str, $MRKR_PDF, "$WORKHTTP/$WORKTEX.pdf");
    &replace_one($str, $MRKR_GIF, "$WORKHTTP/$WORKTEX.gif");
    &replace_one($str, $MRKR_HTML_LATIN1, "$WORKHTTP/$WORKLAT1.htm");
    &replace_one($str, $MRKR_HTML_UTF8, "$WORKHTTP/$WORKUTF8.htm");
    &replace_one($str, $MRKR_POSTSCRIPT, "$WORKHTTP/$WORKTEX.ps");
    &replace_one($str, $MRKR_MSGFILE, "$WORKHTTP/$MSGFILENAME");
    &replace_one($str, $MRKR_TIMETAKEN, "$time_taken");
    &replace_one($str, $MRKR_DONTREDIRECT, "$dontredirect");
    return $str;
}

#----------------------------------------
sub print_cgihtml {
    my($str) = @_;
    # print lines in the CGI output file... until $str is encountered
    # when completed this will finish the
    # body, and do the javascript redirect to webdone.html
    while (<WEBWAIT_THTML>) {
	last if ($str && /$str/);

	if ($errflag) {
	  # depending on error, uncomment particular lines in WEBDONE_THTML
	  next if (&check_marker($_, $MRKR_ANYERROR));
	  if ($ALLCMDS[$errcmd] =~ /$ITRANS_E/) {
	    next if (&check_marker($_, $MRKR_ITRANSERROR));
	  } elsif ($ALLCMDS[$errcmd] =~ /$PGM2GIF_E/) {
	    next if (&check_marker($_, $MRKR_GIFERROR));
	  }
	  next if (&check_marker($_, $MRKR_SYSERROR));
	  # $errcmd not ITRANS or PGMGIF
	} else { 
	  # no error, uncomment the SUCCESS lines
	  next if (&check_marker($_, $MRKR_SUCCESS));
	  # no error, uncomment the LATEXGIF line if latex/pdf/ps/gif produced
	  next if (!$html_only && &check_marker($_, $MRKR_LATEXGIF));
	}

	print &replace_markers($_);
    }
    # this completes the output HTML file (no need to print CGI::end_html)
}
#----------------------------------------
sub sighandler {

    my ($str);

    unlink("$WORKTEX.tex");
    unlink("$WORKTEX.aux", "$WORKTEX.dvi", "$WORKTEX.log", "$WORKTEX.pgm");

    $datestr = localtime(time());
    $str = "$WORKDIRNAME itrans [$datestr] user terminated --\n";
    print LOGFILE $str if (defined fileno LOGFILE);
    print MSGFILE $str if (defined fileno MSGFILE);

    close(WEBDONE_HTML);
    close(WEBWAIT_THTML);
    close(WEBDONE_THTML);
    close(LOGFILE);
    close(MSGFILE);

    exit($errflag);
}
#----------------------------------------

sub myexit {
    # if myexit() is called after WEBWAIT_THTML copied to STDOUT, 
    # the automatic redirect in that file needs to be suppressed
    $dontredirect = "1";

    $errflag++;

    my($mesg) = @_;

    print "<P>\nSession ID# <B>$WORKDIRNAME</B> - Error - $mesg\n<P>";
    # displayed in middle of WEBWAIT_THTML

    $datestr = localtime(time());
    $str = "$WORKDIRNAME [$datestr] Error - $mesg\n";

    print STDERR $str; # goes to Apache error_log or LOGFILE
    print MSGFILE $str if (defined fileno MSGFILE);

    # complete rest of HTML from this CGI script
    &print_cgihtml("");

    close(WEBDONE_HTML);
    close(WEBWAIT_THTML);
    close(WEBDONE_THTML);
    close(LOGFILE);
    close(MSGFILE);

    exit($errflag);
}

#----------------------------------------
sub find_tags {
  my ($line, $lineno) = @_;
  my ($script, $size);
  %scriptsizes = ();
  REMATCH:
  {
    if ( $line =~ /\G\{(\w+)-(\w+)\}/gc ) { # found {script-size}

	# look for {word-word} patterns
	$script = $1;
	$size = $2;

	print MSGFILE "$0: line $lineno loop ($1-$2) found\n" if ($DEBUG >= 2);
	if (! grep(/^$script$/, @SCRIPTS) || ! defined $FONTSIZES{$size} ) {
	    print MSGFILE "Warning - line $lineno - unknown script-fontsize found: \"$script-$size\", skipping.\n";
	} else {
	    # using hash as a set data structure
	    $scriptsizes{"$script-$size"} = 1;
	}

	redo REMATCH;
    } # end of if() found script-size

    if ( $line =~ /\G.[^{]*/gc ) {
	# first dot consumes one char ( may be a { )
	# then ignore all characters other than another {
	redo REMATCH;
    } # end of if() found script-size
  } # end REMATCH block that contains if()

  %scriptsizes;
}

#----------------------------------------
# Try a few times to make unique temp directory by appending chars if needed

sub mktempdir {
    my $dir = shift; # directory to create tmp dir in
    my $try = shift; # prefix to use for directory name
    my $n = shift; # a numeric suffix to use - is incremented until unique
    my $done = 0;
    my $i;
    my $dirname;

    umask(0);
    foreach $i (0..9) {
	$dirname = "$try-$n";
	$done = mkdir("$dir/$dirname", 0777);
	last if ($done);
	$n++;
    }
    $done || myexit("mktempdir: mkdir in ($dir) with prefix ($try-$n) failed: $!");
    $done ? $dirname : "";
}

#----------------------------------------
sub initialize {

    umask(0);

    select((select(STDOUT), $| = 1)[0]);

    # output HTML page headers, rest of HTML is output by copying WEBWAIT_THTML
    print CGI::header();

    open(LOGFILE, ">> $LOGFILE")  
      or myexit("Could not open log file ($LOGFILE): $!");
    select((select(LOGFILE), $| = 1)[0]);

    # web users come in as same userid as the script is run on,
    # so to prevent them from reading the log file (small measure of
    # security), make it not-readable by this owner, assuming that
    # the file was created by this script!
    if (-r $LOGFILE && -o $LOGFILE) {
	chmod 0226, $LOGFILE
	  or print LOGFILE "Warning: Could not chmod log file: $!";
    }

    # all STDERR output is caught in local log file, so any perl errors
    # or uncaught STDERR from commands goes here, instead of the global
    # apache log file
    open (STDERR, '>>& LOGFILE')
      or myexit("Could not redirect STDERR to log file ($LOGFILE): $!");
}

#-----------------------------------------------------------------------------
